from paracrine.debian import apt_install, debian_repo
from paracrine.fs import build_with_command, download
from paracrine.systemd import systemd_set


def core_run():
    apt_install(["curl"])
    download(
        "https://www.postgresql.org/media/keys/ACCC4CF8.asc",
        "/etc/apt/trusted.gpg.d/postgresql.gpg.asc",
        "0144068502a1eddd2a0280ede10ef607d1ec592ce819940991203941564e8e76",
    )
    apt_install(["gpg"])
    build_with_command(
        "/etc/apt/trusted.gpg.d/postgresql.gpg",
        "cat /etc/apt/trusted.gpg.d/postgresql.gpg.asc | gpg --dearmor > /etc/apt/trusted.gpg.d/postgresql.gpg",
        ["/etc/apt/trusted.gpg.d/postgresql.gpg.asc"],
    )
    debian_repo(
        "postgresql_org_repository",
        "deb http://apt.postgresql.org/pub/repos/apt bullseye-pgdg main",
    )
    apt_install(["postgresql-14"])
    systemd_set("postgresql", enabled=True, running=True)



import os
from typing import List

from .fs import run_command, set_file_contents


def apt_install(packages: List[str], always_install: bool = False) -> None:
    if not always_install:
        packages = [
            package
            for package in packages
            if not os.path.exists(f"/var/lib/dpkg/info/{package}.list")
            and not os.path.exists(f"/var/lib/dpkg/info/{package}:amd64.list")
        ]
        if packages == []:
            return
    # Confdef is to fix https://unix.stackexchange.com/a/416816/73838
    os.environ["DEBIAN_FRONTEND"] = "noninteractive"
    run_command(
        "apt-get install %s --no-install-recommends --yes -o DPkg::Options::=--force-confdef"
        % " ".join(packages)
    )


def apt_update():
    run_command("apt-get update --allow-releaseinfo-change")


def debian_repo(name, contents=None):
    fname = "/etc/apt/sources.list.d/%s.list" % name
    if contents is None:
        contents = "deb http://deb.debian.org/debian %s main" % name

    changed = set_file_contents(fname, contents)
    if changed:
        apt_update()



import json
import os
import pathlib
from typing import Any, Dict, Optional

import jinja2
import yaml

_jinja_env = None
data = None

CONFIG_NAME = "config.yaml"


def jinja_env():
    assert _jinja_env is not None, "Need to run set_data first!"
    return _jinja_env


def host():
    return data["host"]


def config():
    return data["config"]


def get_config_file(fname):
    return data["configs"][fname]


def core_config():
    return yaml.safe_load(get_config_file(CONFIG_NAME))


def set_data(new_data: Dict[str, Any]) -> None:
    loader = jinja2.DictLoader(new_data["templates"])
    global _jinja_env, data
    _jinja_env = jinja2.Environment(
        loader=loader, undefined=jinja2.StrictUndefined, keep_trailing_newline=True
    )
    data = new_data


def add_folder_to_config(configs, folder, shortname=None, filter=None):
    if not os.path.exists(folder):
        print("Skipping %s from config as doesn't exist" % folder)
        return
    for f in os.listdir(folder):
        if filter is not None:
            if not filter(f):
                continue
        if shortname is not None:
            key = os.path.join(shortname, f)
            full = os.path.join(folder, f)
        else:
            full = key = os.path.join(folder, f)
        configs[key] = open(full).read()


inventory = None
inventory_directory = None


def set_config(inventory_path):
    global inventory, inventory_directory
    inventory_directory = os.path.dirname(inventory_path)
    inventory = yaml.safe_load(open(inventory_path))


def get_config():
    if inventory is None:
        assert data is not None
        return data["inventory"]
    return inventory


def config_path(shortname=False):
    if inventory is None or shortname:
        return "configs"
    else:
        return os.path.join(data_path(), "configs")


def path_to_config_file(name: str) -> str:
    if name.startswith("/"):
        return name
    return os.path.normpath(f"{config_path()}/{name}")


def data_path():
    if inventory is None:
        return None
    return os.path.join(inventory_directory, inventory["data_path"])


def environment():
    if inventory is None:
        return data["environment"]
    return inventory["environment"]


def create_data(server: Optional[Dict] = None):
    config = get_config()
    templates = {}
    template_paths = [
        pathlib.Path("templates"),
        pathlib.Path(__file__).parent.joinpath("templates"),
    ]
    for template_path in template_paths:
        if not template_path.exists():
            continue
        for path in template_path.iterdir():
            templates[path.name] = path.open().read()

    configs = {
        "config.yaml": open("config.yaml").read(),
    }
    add_folder_to_config(
        configs,
        config_path(),
        shortname="configs",
        filter=lambda f: f.startswith("wireguard-public")
        or f.startswith("networks-")
        or f.startswith("other-")
        or f.startswith("ssh-"),
    )

    return {
        "templates": templates,
        "host": server,
        "config": config,
        "configs": configs,
        "environment": environment(),
        "inventory": get_config(),
    }


def network_config_file(name, shortname=False):
    return config_path(shortname) + "/networks-{name}".format(name=name)


def network_config(name):
    return json.loads(get_config_file(network_config_file(name, shortname=True)))


def other_config_file(name, shortname=False):
    return config_path(shortname) + "/other-{name}".format(name=name)


def other_config(name):
    return json.loads(get_config_file(other_config_file(name, shortname=True)))


def other_self_config():
    return other_config(host()["name"])


def build_config(config: Dict) -> Dict:
    env = environment()
    LOCAL = config["environments"][env]
    common = config.get("common", {})
    ret = dict(**common)
    ret.update(LOCAL)
    return ret


def local_config() -> Dict:
    return yaml.safe_load(open(CONFIG_NAME).read())



import pathlib

from .debian import apt_install
from .fs import run_command


def setup_venv(venv: pathlib.Path) -> None:
    venv_bin = venv.joinpath("bin")
    venv_pip = venv_bin.joinpath("pip")

    if not venv_pip.exists():
        apt_install(["python3-venv"])
        run_command(f"python3 -m venv {venv}")

    if not venv_bin.joinpath("wheel").exists():
        run_command(f"{venv_pip} install wheel")



import json
import os
import sys
from distutils.version import LooseVersion

from ..config import (
    config,
    config_path,
    get_config,
    get_config_file,
    host,
    network_config,
    other_config_file,
)
from ..core import in_docker
from ..debian import apt_install
from ..fs import make_directory, run_command, set_file_contents_from_template
from ..network import external_ip
from ..systemd import systemd_set

wg_config = "/etc/wireguard"
private_key_file = "{wg_config}/privatekey".format(wg_config=wg_config)
public_key_file = "{wg_config}/publickey".format(wg_config=wg_config)


def public_key_path(name):
    return config_path() + "/wireguard-public-{name}.key".format(name=name)


def get_output(host, command):
    status, stdout, stderr = host.run_shell_command(command=command)
    stdout = "".join(stdout)
    assert status is True, (stdout, stderr)
    return stdout


def get_all_kernel_versions():
    raw = run_command(
        r"dpkg-query --showformat=$\{Package\},$\{Status\},$\{Version\}\\t --show linux-image-*",
    )
    versions = {}
    for line in raw.split("\t"):
        if line.strip() == "":
            continue
        try:
            (name, status, version) = line.split(",")
        except ValueError:
            raise Exception("'%s'" % line)
        if status == "install ok installed":
            versions[name] = version
    return versions


def bootstrap_run():
    apt_install(["kmod"])
    modules = sorted([line.split(" ")[0] for line in run_command("lsmod").splitlines()])
    if "wireguard" not in modules:
        print("modules", modules)
        apt_install(["linux-image-amd64"])
        versions = get_all_kernel_versions()
        ordered = sorted(
            [
                x.replace("linux-image-", "")
                for x in versions.keys()
                if x not in ["linux-image-amd64", "linux-image-cloud-amd64"]
            ],
            key=LooseVersion,
            reverse=True,
        )
        highest = ordered[0]
        current = run_command("uname -r").strip()
        if current != highest:
            print(ordered)
            print("'%s' != '%s'" % (current, highest))
            apt_install(["systemd-sysv"])
            if not in_docker():
                run_command("reboot")
                sys.exit(0)

        apt_install(["wireguard"])
        if not in_docker():
            apt_install(["linux-headers-amd64"])
            modules = sorted(
                [line.split(" ")[0] for line in run_command("lsmod").splitlines()]
            )
            if "wireguard" not in modules:
                print("modules", modules)
                run_command("modprobe wireguard")

    make_directory(wg_config)
    if not os.path.exists(private_key_file):
        run_command("wg genkey > %s" % private_key_file)
    if not os.path.exists(public_key_file):
        run_command("cat %s | wg pubkey > %s" % (private_key_file, public_key_file))

    return {"wg_publickey": open(public_key_file).read().strip()}


def bootstrap_parse_return(info):
    open(public_key_path(host()["name"]), "w").write(info["wg_publickey"])

    wg_ips = []

    for server in get_config()["servers"]:
        networks = network_config(server["name"])
        wireguard_networks = [
            network for network in networks if network["ifname"] == "wg0"
        ]
        if len(wireguard_networks) == 1:
            wg_ips.append(wireguard_networks[0]["addr_info"][0]["local"])
    json.dump(wg_ips, open(other_config_file("wireguard-ips"), "w"), indent=2)


def setup(name="wg0", ip="192.168.2.1", netmask=24, peers=[]):
    peers = {}
    for h in config()["servers"]:
        if h["name"] == host()["name"]:
            continue
        public_key = get_config_file(public_key_path(h["name"]))
        peers[h["name"]] = {
            "public_key": public_key,
            "endpoint": "%s:51820" % external_ip(h),
            "peer_addr": h["wireguard_ip"],
        }

    conf_change = set_file_contents_from_template(
        f"/etc/wireguard/{name}.conf",
        "wg.conf.j2",
        PRIVATE_KEY=open(private_key_file).read().strip(),
        PEERS=peers,
        IP=ip,
        NETMASK=netmask,
    )

    systemd_set(f"wg-quick@{name}", enabled=True, restart=conf_change)


def core_run():
    setup(ip=host()["wireguard_ip"])


__all__ = ["core_run", "bootstrap_run", "bootstrap_parse_return"]



import importlib
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from .config import set_data

Modules = List[Union[ModuleType, Tuple[ModuleType, Dict]]]
"""Type of modules handed to `paracrine.runner.run`"""

TransmitModules = List[Union[str, Tuple[str, Dict]]]


def runfunc(
    modules: Modules, name: str, arguments: Dict[str, Any] = {}, data: Dict = {}
) -> Dict[str, Any]:
    ret = {}

    def run(module: ModuleType, options: Dict):
        func: Optional[Callable] = getattr(module, name, None)
        if func is not None:
            setattr(module, "options", options)
            if data != {}:
                set_data(data)
            try:
                if module.__name__ in arguments:
                    ret[module.__name__] = func(arguments[module.__name__])
                else:
                    ret[module.__name__] = func()
            except Exception:
                print(f"Error while running {name} for {module.__name__}")
                raise

    for module in modules:
        if isinstance(module, ModuleType):
            run(module, {})
        else:
            run(module[0], module[1])
    return ret


def maketransmit(modules: Modules) -> TransmitModules:
    ret = []
    for module in modules:
        if isinstance(module, ModuleType):
            ret.append(module.__name__)
        else:
            ret.append((module[0].__name__, module[1]))

    return ret


def makereal(modules: TransmitModules) -> Modules:
    ret = []
    for module in modules:
        if isinstance(module, str):
            ret.append(importlib.import_module(module))
        else:
            ret.append((importlib.import_module(module[0]), module[1]))

    return ret



from mitogen.parent import Router
from mitogen.utils import run_with_router

from paracrine.config import set_config
from paracrine.fs import run_command
from paracrine.runner import main


def do(data):
    run_command("reboot")


def core(router: Router) -> None:
    for _ in main(router, do):
        pass


if __name__ == "__main__":
    import sys

    set_config(sys.argv[1])
    run_with_router(core)
